

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Accumulate_and_return Example Program using the the LAPI_Put api
**   call.  For a set of n tasks 0,1,...,n-1, where n is an even
**   number, all tasks are divided into (src,tgt) buddy pairs (0,1),
**   (2,3), etc.  For each pair, an array of ints is transferred
**   from src to tgt with a LAPI_Put call.  The ints are then
**   manipulated and sent back to the src with another LAPI_Put
**   call.  All tasks wait on target counters to signal
**   completion of message delivery.  The various data buffers are
**   printed along the way to show progress.
*/


/* stores initial value on src and final value on tgt */
int           data_buffer[ARRAYLEN];

/* store list of remote buffer addrs */
void          **data_buffer_list;  /* remote data buffer addrs */

/* store list of remote tgt_cntr addrs */
void          **tgt_cntr_list;     /* remote tgt cntr addrs    */

/* updates on tgt at msg completion */
lapi_cntr_t   tgt_cntr;


int main( int argc, char **argv )
{

lapi_handle_t handle;        /* LAPI handle */
lapi_info_t   info;          /* Info to pass to LAPI_Init */
int           task_id;       /* Our LAPI Task id */
int           num_tasks;     /* Total number of tasks */
int           i,j;           /* Loop counters */
int           buddy;         /* Our communication partner */
int           val;           /* Needed for waitcntr call on source */

    /* Not passing any info to init through this struct. */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks < 2) || ((num_tasks % 2) != 0) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }

    /* allocate space for remote addrs */
    data_buffer_list = (void **) malloc(num_tasks*sizeof(void *));
    tgt_cntr_list = (void **) malloc(num_tasks*sizeof(void *));

    /* each task executes on tgt_cntr within its own data space */
    CHECK((LAPI_Setcntr(handle,&tgt_cntr,0)));

    /*
    ** collective calls.  each task stores its own addr plus remote
    ** addrs
    */
    CHECK((LAPI_Address_init(handle,
                           &tgt_cntr,tgt_cntr_list)));
    CHECK((LAPI_Address_init(handle,
                                        (void *)data_buffer,
                                         data_buffer_list)));

    /*
    ** up to this point, all instructions have executed on all
    ** tasks. we now begin differentiating tasks.
    */
    if ( task_id % 2  == 0 ){   /* sender */
        buddy = task_id + 1;

        /* initialize data buffer */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = i;
        }

        for (i = 0; i < ARRAYLEN; i++ ) {
            printf("original data_buffer[%d]: %d\n", i, data_buffer[i]);
        }

        /* collective call.  sync before starting data transfer */
        CHECK((LAPI_Gfence(handle)));

        /*
        ** execute the data transfer to our buddy task.  send
        ** ARRAYLEN ints, starting with data_buffer[0].  Note that
        ** no user header is sent, and the target counter is used
        ** to signal message completion at the target.
        ** args:  handle                  -- the LAPI handle
        **        buddy                   -- the target task id
        **        ARRAYLEN*sizeof(int)    -- the length of data to
        **                                   transfer
        **        data_buffer_list[buddy] -- remote addr for
        **                                   writing data
        **        &(data_buffer[0])       -- the starting address
        **                                   of data to transfer
        **        tgt_cntr_list[buddy]    -- remote tgt counter
        **                                   address.  will update
        **                                   when message completes
        **                                   on target.
        **        NULL, NULL              -- origin and completion
        **                                   counters.
        */

        CHECK((LAPI_Put(handle, buddy,
                           (ulong) (ARRAYLEN*sizeof(int)), 
                           (void *) (data_buffer_list[buddy]),
                           (void *) (&(data_buffer[0])), 
                           (void *)(tgt_cntr_list[buddy]),
                           NULL, NULL)));  

        /* now wait for tgt to return data to us */
        CHECK((LAPI_Waitcntr(handle, &tgt_cntr,
                                         1, &val)));
        for (i = 0; i < ARRAYLEN; i++ ) {
            printf("final data_buffer[%d]: %d\n", i,
                            data_buffer[i]);
        }

    } else {             /* receiver */
        int tmp_buf[ARRAYLEN]; /* for intermediate values */

        buddy = task_id -1;

        for( j = 0; j < ARRAYLEN; j++ ) {
            tmp_buf[j] = ARRAYLEN - (2*j);
        }

        /* match src's sync */
        CHECK((LAPI_Gfence(handle)));

        /* wait for data to arrive from src */
        CHECK((LAPI_Waitcntr(handle, &tgt_cntr,1,
                           &val)));

        /* store different set of values in array */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] += tmp_buf[i];
        }

        /*
        ** Send new data values back to our buddy.  Note the use
        ** of the buddy entries in the data_buffer_list and
        ** tgt_cntr_list for remote addresses.
        */
        CHECK((LAPI_Put(handle, buddy,
                           (ulong) (ARRAYLEN*sizeof(int)), 
                           (void *) (data_buffer_list[buddy]),
                           (void *) (&(data_buffer[0])), 
                           (void *) tgt_cntr_list[buddy],
                           NULL, NULL)));  
    }

    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(data_buffer_list);
    free(tgt_cntr_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


