

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Example Program to illustrate the use of the
**   LAPI_Xfer interface for DGSP data transfers.
**   For a set of n tasks 0,1,...,n-1, where n is
**   an even number, all tasks are divided into
**   (src,tgt) buddy pairs (0,1), (2,3), etc.   A simple
**   DGSP program is created that ships alternatei
**   integers from a data buffer on the src.  The
**   receive side places these integers into a
**   contiguous array.  The result is an array on the
**   target that is half the length of the array on the
**   src.
**
**   The gather is accomplished by using a DGSP with one
**   copy and one iterate instruction.  The copy has a
**   size of 4 bytes and an extent of 8 bytes.  This
**   causes the machine to copy 4, skip 8, copy 4,
**   etc...until the amount specified by the user 
**   is transferred.  
**
**   The src of each pair creates a data buffer of
**   ints then prints the values.  A source-side DGSP
**   is built as described above.  the DGSP is 
**   registered using LAPI_Util then run using 
**   LAPI_Xfer.  As the DGSP runs, LAPI iterates through
**   the array and transfers alternate elements as
**   described above.  The receive completion handler
**   verifies the results of the Xfer call by printing
**   the contents of the target-side data buffer.
**
**   LAPI will automatically invoke the completion handler
**   upon message completion.  Note that the origin task
**   must wait on the completion counter before terminating,
** to ensure that the completion handler has executed.
*/

#define  DOUBLE_ARRAYLEN ARRAYLEN*2

/*
** stores initial value on src and final value on tgt
** note that DOUBLE_ARRAYLEN size is needed on the
** src to get ARRAYLEN ints on the tgt in this
** example.
*/
int           data_buffer[DOUBLE_ARRAYLEN];


/* global array of header handler addrs */
lapi_long_t   *hdr_hdl_list;

/*
** DGSP code array.  DGSP instructions are stored
** as ints (with constants defined in lapi.h for
** the number of ints needed to store each
** instruction).  We will have one COPY and one ITERATE
** instruction in our DGSP.  We use LAPI's constants
** to allocate the appropriate storage.
*/
int  code[LAPI_DGSM_COPY_SIZE+LAPI_DGSM_ITERATE_SIZE];

/* DGSP descriptor */
lapi_dgsp_descr_t dgsp_d;

/* updates on src at msg completion */
lapi_cntr_t   compl_cntr;

/*
**  completion handler that runs on target after
**  completion of message delivery.
**  simply prints the contents of the data buffer that
**  results from the DGSP xfer.
*/
void compl_hndlr(lapi_handle_t *handle, void *param)
{
    int i;              /* loop counter */
    
    printf("Greetings from the completion handler...\n");
    for( i = 0; i < (ARRAYLEN); i++ ) {
        printf("final buffer[%d]: %d\n",i, data_buffer[i]);
    }

}

/* header handler that runs on target when first packet
** arrives.
**
** sets the completion_handler pointer and returns the
** address of the tgt side data buffer.  Since only
** an address is returned (and not a DGSP), LAPI simply
** writes the gathered data as a contiguous stream 
** (similar to how data is written in an Amsend call).
** Had the header handler returned a pointer to a
** DGSP descriptor, LAPI would have interpreted the
** DGSP to unpack the data at the target.
*/
void *header_handler(lapi_handle_t *handle, void *uhdr,
                     uint uhdr_len, uint msg_len,
                     compl_hndlr_t **completion_handler,
                     void **user_info)
{
    /* pass the completion handler back by reference */
    *completion_handler = compl_hndlr; 
    printf("Greetings from the header_handler, returning:0x%x\n",data_buffer);
    return data_buffer;
}


int main( int argc, char **argv )
{

lapi_handle_t handle;    /* LAPI handle */
lapi_info_t   info;      /* Info to pass to LAPI_Init */
int           task_id;   /* Our LAPI Task id */ 
int           val;       /* Needed for waitcntr call */
int           num_tasks; /* Total number of tasks */

/*
** Table of remote data buf addrs
*/
void          *buf_addr_list[NUM_TASKS];
int           i,k;       /* Loop counter */
int           buddy;     /* Our communication partner */

/*
** Data structure for the xfer call.
*/
lapi_xfer_t   xfer_struct;


/* DGSP data structures */
lapi_dgsm_copy_t    *copy_p; /* copy instruction */
lapi_dgsm_iterate_t *iter_p; /* iterate instruction */
int                 *code_ptr; /* code pointer */

/* constant for holding code array info */
int                 code_less_iterate_size;

/* used for DGSP registration */
lapi_reg_dgsp_t     reg_util;


    /*
    ** Clear the struct.  Not passing any info to init
    ** through this struct.
    */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id))); 

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks < 2) || ((num_tasks % 2) != 0) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }  
    hdr_hdl_list = (lapi_long_t *)
               malloc(num_tasks*sizeof(lapi_long_t));

    /*
    ** collective calls.  each task stores its own addr
    ** plus remote addrs
    */
    CHECK((LAPI_Address_init64(handle,(lapi_long_t)&header_handler,
                               hdr_hdl_list)));

    /*
    ** up to this point, all instructions have executed
    ** on all tasks.
    ** we now begin differentiating tasks.
    */
    if ((task_id %2) == 0) {   /* message source */

       /* setup buddy pairs as (0,1), (2,3), etc. */
        buddy = task_id+1;

       /*
       ** Set up dgsp descriptor
       */

       /* set pointer to code array */
       dgsp_d.code = &code[0];

       /* set size of code array */
       dgsp_d.code_size = LAPI_DGSM_COPY_SIZE +
                          LAPI_DGSM_ITERATE_SIZE;

       /* not using DGSP gosub instruction */
       dgsp_d.depth = 1;

       /* 
       ** set density to show internal gaps in the
       ** DGSP data layout
       */
       dgsp_d.density   = LAPI_DGSM_SPARSE;

       /* transfer 4 bytes at a time */
       dgsp_d.size  = 4;

       /* advance the template by 8 for each iteration */
       dgsp_d.extent = 8;

       /*
       ** ext specifies the memory 'footprint' of
       ** data to be transferred. The lext specifies
       ** the offset from the base address to begin
       ** vieiwing the data.  The rext specifies the
       ** length from the base address to use.
       */
       dgsp_d.lext = 0;
       dgsp_d.rext = 4;

       /* atom size of 0 lets LAPI choose the pkt size */
       dgsp_d.atom_size = 0;

       /* 
       ** setup the copy instruction
       */
       copy_p = (lapi_dgsm_copy_t *)(dgsp_d.code);
       copy_p->opcode = LAPI_DGSM_COPY;

       /* copy 4 bytes at a time */      
       copy_p->bytes = (long) 4;

       /* start at offset 0 */
       copy_p->offset = (long) 0;

       /* set code ptr to addr of iterate instruction */
       code_less_iterate_size =  dgsp_d.code_size - LAPI_DGSM_ITERATE_SIZE;
       code_ptr = ((int *)(code))+code_less_iterate_size;

       /* 
       ** Setup iterate instruction
       */
       iter_p = (lapi_dgsm_iterate_t *) code_ptr;
       iter_p->opcode = LAPI_DGSM_ITERATE;
       iter_p->iter_loc = (-code_less_iterate_size);

       /* Setup and do DGSP registration */
       reg_util.Util_type = LAPI_REGISTER_DGSP;
       reg_util.idgsp = &dgsp_d;
       CHECK((LAPI_Util(handle,
                         (lapi_util_t *)&reg_util)));


    /* initialize data buffer. */
        for (i = 0; i < (DOUBLE_ARRAYLEN); i++ ) {
            data_buffer[i] = i*buddy;
        }

       /*
       ** collective call.  sync before starting data
       ** transfer
       */
       CHECK((LAPI_Gfence(handle)));

    /*
    ** initialize the completion counter. When it increments, we know
    ** that the completion handler has returned
    */
    CHECK((LAPI_Setcntr(handle,&compl_cntr,0)));

   /* 
   ** Setup the xfer struct to transfer to our buddy
   ** task.  start at data_buffer[0], send one int,
   ** skip 8 bytes, send the next int, skip 8 bytes,
   ** etc. starting with data_buffer[0].  Note that no
   ** user header is sent, and no counters are used.
   ** args:  handle                 -- the LAPI handle
   **        buddy                  -- the target task id
   **        header_handler         -- the header handler
   **                                  that will execute
   **                                  on the target
   **        NULL                   -- the user header 
   **        0                      -- the user header
   **                                  length
   **        &(data_buffer[0])      -- the starting address
   **                                  of data to transfer
   **        ARRAYLEN*sizeof(int)   -- the length of data
   **                                  to transfer
   **        NULL, NULL             -- target and origin counter
   **        compl_cntr             -- completion counter
   */
        /* required for DGSP transfer */
        xfer_struct.Dgsp.Xfer_type = LAPI_DGSP_XFER;

        /* no flags sent to LAPI */
        xfer_struct.Dgsp.flags = 0;

        /* set data target */
        xfer_struct.Dgsp.tgt = buddy;

        /* set header handle to remote addr */
        xfer_struct.Dgsp.hdr_hdl = (lapi_long_t)
                           (hdr_hdl_list[buddy]);

        /* not using user header */
        xfer_struct.Dgsp.uhdr = (void *) NULL;
        xfer_struct.Dgsp.uhdr_len = 0;

        /* set pointer to data base address */
        xfer_struct.Dgsp.udata = (void *) data_buffer;

        /* tell LAPI to transfer ARRAYLEN ints */
        xfer_struct.Dgsp.udata_len = ARRAYLEN*(sizeof(int));
        /* not using send completion handler */
        xfer_struct.Dgsp.shdlr = (scompl_hndlr_t *) NULL;
        xfer_struct.Dgsp.sinfo = (void *) 0;

        /* not using counters */
        xfer_struct.Dgsp.org_cntr = NULL;
        xfer_struct.Dgsp.cmpl_cntr = &compl_cntr;
        xfer_struct.Dgsp.tgt_cntr = (lapi_long_t) NULL;

        /*
        ** set dgsp handle to the one returned by LAPI
        ** at DGSP registration (LAPI_Util call above).
        */
        xfer_struct.Dgsp.dgsp = reg_util.dgsp_handle;

        /* do the xfer */
        CHECK((LAPI_Xfer(handle, &xfer_struct)));

        /* wait for message completion before final termination */
        CHECK((LAPI_Waitcntr(handle, &compl_cntr,1, &val)));

     } else {             /* receiver */
        /* to match the src's sync before data transfer */
        CHECK((LAPI_Gfence(handle)));
    }

    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(hdr_hdl_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


