

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Example Program to illustrate the use of various LAPI Address
** routines, namely LAPI_Address_init, LAPI_Address_init64,
** LAPI_Addr_set  and LAPI_Addr_get. For a set of n tasks
** 0,1,...,n-1, where n is an even number, each of the above
** routines is called, and the resulting data is printed.  
**
*/

/* some local pointers whose addresses we will exchange */
void          *local_addr1, *local_addr2;

/* list of addresses gathered from all tasks */
void          **addr_list;
lapi_long_t   *long_addr_list;   /* needed for LAPI_Address_init64 */


int main( int argc, char **argv )
{
    lapi_handle_t handle;        /* LAPI handle */
    lapi_info_t   info;          /* Info to pass to LAPI_Init */
    int           task_id;       /* Our LAPI Task id */
    int           num_tasks;     /* Total number of tasks */
    int           i,j;           /* Loop counters */
    void          *tbl_addr;     /* For result of Addr_get call */

    /* Not passing any info to init through this struct. */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* initialize the table buffers for the data addrs. */
    addr_list = (void **) malloc(num_tasks*sizeof(void *));
    long_addr_list = (lapi_long_t *) malloc(num_tasks * sizeof(lapi_long_t));

    /* collective call.  each task stores its own addr plus remote addrs */
    CHECK((LAPI_Address_init(handle, (void *) &local_addr1,
                                         addr_list)));
    CHECK((LAPI_Address_init64(handle, (lapi_long_t) &local_addr2,
                                           long_addr_list)));

    /* show local address, then list table of remote addresses */
    printf("address of local_addr1: 0x%x\n", task_id, &local_addr1);

    for( i=0; i < num_tasks; i++ ) {
        printf("addr_list[%d]: 0x%x\n", i, addr_list[i]);
    }

    printf("address of local_addr2 (as lapi_long_t): 0x%llx\n", 
           task_id, (lapi_long_t) &local_addr2);

    for( j=0; j < num_tasks; j++ ) {
        printf("long_addr_list[%d]: 0x%llx\n", j, long_addr_list[j]);
    }

    printf("Storing address 0x%x at position 1.\n", &local_addr1);
    CHECK((LAPI_Addr_set(handle, (void *)&local_addr1, 1)));

    CHECK((LAPI_Addr_get(handle, (void *)&tbl_addr, 1)));
    printf("Retrieved address 0x%x from position 1.\n", tbl_addr);

    /* sync before cleanup */
    CHECK((LAPI_Gfence(handle)));


    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(addr_list);
    free(long_addr_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


