

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Example Program to illustrate the use of the LAPI_Amsend api call.
**   For a set of n tasks 0,1,...,n-1, where n is an even number, all
**   tasks are divided into (src,tgt) buddy pairs (0,1), (2,3), etc.
**   For each pair, an array of ints is transferred from src to tgt
**   with a single LAPI_Amsend call.  The receive completion handler
**   verifies the results of the Amsend call by printing the contents
**   of the target-side data buffer.  Note that the completion counter
**   is used to ensure that the target side completion handler has
**   finished execution before the src side terminates.
*/

/* list of header handler addrs */
void **hdr_hndlr_list;

/* stores initial value on src and final value on tgt */
int           data_buffer[ARRAYLEN];

/* updates on src at msg completion */
lapi_cntr_t   compl_cntr;

/*
**  completion handler that runs on target after completion of message delivery.
**    simply prints the contents of the data buffer that is the
**    target of the Amsend call
*/
void compl_hndlr(lapi_handle_t *handle, void *param)
{
    int i;              /* loop counter */
    
    printf("Greetings from the completion handler...\n");
    for( i = 0; i < ARRAYLEN; i++ ) {
        printf("final buffer[%d]: %d\n",i, data_buffer[i]);
    }

}

/* header handler that runs on target when first packet arrives.
**   sets the completion_handler pointer and returns the address
**   of the data buffer for message delivery.  LAPI writes the
**   Amsend data at this address.
*/
void *header_handler(lapi_handle_t *handle, void *uhdr, uint uhdr_len, uint msg_len,
                      compl_hndlr_t **completion_handler, void **user_info) {

    /* pass the completion handler back by reference */
    *completion_handler = compl_hndlr; 

    return data_buffer;
}


int main( int argc, char **argv )
{

lapi_handle_t handle;                      /* LAPI handle */
lapi_info_t   info;                        /* Info to pass to
                                              LAPI_Init */
int           task_id;                     /* Our LAPI Task id */ 
int           num_tasks;                   /* Total number of tasks */
int           i;                           /* Loop counter */
int           val;                         /* Needed for waitcntr call */
int           buddy;                       /* Our communication
                                              partner */

    /*
    ** Clear the struct.  Not passing any info to init through this
    ** struct.
    */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id))); 

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks < 2) || ((num_tasks % 2) != 0) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }  

    /* get address of header handler on tgt */
    hdr_hndlr_list = (void **)malloc(num_tasks*(sizeof(void *)));
    CHECK((LAPI_Address_init(handle,(void *)header_handler,
                                         hdr_hndlr_list)));

    /*
    ** up to this point, all instructions have executed on all 
    ** tasks. we now begin differentiating tasks.
    */
    if ((task_id %2) == 0) {   /* message source */

    /* setup buddy pairs as (0,1), (2,3), etc. */
        buddy = task_id+1;

    /* initialize data buffer. */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = i*buddy;
        }

        /*
        ** initialize the completion counter. When it increments, we know
        ** that the completion handler has returned
        */
        CHECK((LAPI_Setcntr(handle,&compl_cntr,0)));

    /* collective call.  sync before starting data transfer */
        CHECK((LAPI_Gfence(handle)));

    /* 
    ** execute the data transfer to our buddy task.  send ARRAYLEN 
    ** ints, starting with data_buffer[0].  Note that no user
    ** header is sent, and no counters are used.
    ** args:  handle                 -- the LAPI handle
    **        buddy                  -- the target task id
    **        header_handler         -- the header handler that will 
    **                                  execute on the target
    **        NULL                   -- the user header 
    **        0                      -- the user header length
    **        &(data_buffer[0])      -- the starting address of data to
    **                                  transfer
    **        ARRAYLEN*sizeof(int)   -- the length of data to transfer
    **        NULL, NULL             -- target and origin counters
    **        compl_cntr             __ completion counter
    */
        CHECK((LAPI_Amsend(handle, buddy,
                           hdr_hndlr_list[buddy], NULL, 
                           0,&(data_buffer[0]),ARRAYLEN*(sizeof(int)),
                           NULL, NULL, &compl_cntr)));

        /* wait for message completion before final termination */
        CHECK((LAPI_Waitcntr(handle, &compl_cntr,1, &val)));

     } else {             /* receiver */
         /* buddy = task_id-1; */

         /* initialize the data buffer (not necessary, but provided for
            clarity) */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = 0;
        }

        /* to match the src's sync before data transfer */
        CHECK((LAPI_Gfence(handle)));
    }

    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(hdr_hndlr_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


