

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

#define VECLEN         15
#define NUM_VECS       10
#define ARRAYLEN       10

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];



/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Example Program to illustrate the use of the LAPI_Amsendv api call.
** For a set of n tasks 0,1,...,n-1, where n is an even number, all
** tasks are divided into (src,tgt) buddy pairs (0,1), (2,3), etc.
** For each pair, an array of vectors is transferred from src to tgt
** with a single LAPI_Amsendv call.  Each vector contains a single int.
** The receive completion handler verifies the results of the Amsend
** call by printing the contents of the target-side data buffer.
** On the source side, the origin counter is used so that the
** malloc'ed vector structure can be freed.  The completion counter
** is used to ensure that the completion handler finishes on the target
** before the src task terminates.
*/

/* list of header handler addrs */
void **hdr_hndlr_list;

/* origin vector on src, target vector on tgt. */
lapi_vec_t *vec;

/* origin counter for sync'ing on source side. */
lapi_cntr_t origin_cntr;

/* updates on src at msg completion */
lapi_cntr_t   compl_cntr;

/*
**  completion handler that runs on target after completion of message 
**  delivery.
**  simply prints the contents of the data buffer that is the
**  target of the Amsendv call
*/
void compl_hndlr(lapi_handle_t *handle, void *param)
{
    int i;              /* loop counter */
    
    printf("Greetings from the completion handler...\n");
    for( i=0; i < NUM_VECS; i++ ) {
        int *int_ptr = (int *)(vec->info[i]);
        printf("vec->info[%d]: %d\n",i,*int_ptr);
    }

    /* free all structs allocated in the header handler */
    for( i=0; i < NUM_VECS; i++ ) {
       free(vec->info[i]);
    }
    free(vec->info);
    free(vec->len);
    free(vec);
}

/* header handler that runs on target when first packet arrives.
**   sets the completion_handler pointer and returns the address
**   of the target vector.
*/
lapi_vec_t *header_handler(lapi_handle_t *handle, void *uhdr, 
                           uint uhdr_len, ulong *len_vec[],
                           compl_hndlr_t **completion_handler,
                           void **user_info)
{
    int i;                      /* loop counter */

    /* initialize the vector whose address will be returned to LAPI */ 
    vec = (lapi_vec_t *) malloc(sizeof(lapi_vec_t));

    /* 
    ** set the fields to match those of the origin vector. allocate storage for the length
    ** and data arrays.  malloc'ed data will be freed in the completion handler.
    */
    vec->num_vecs = NUM_VECS;
    vec->vec_type = LAPI_GEN_IOVECTOR;
    vec->len = (unsigned long *)malloc(NUM_VECS*sizeof(unsigned long));
    vec->info = (void **) malloc(NUM_VECS*sizeof(void *));
    for( i=0; i < NUM_VECS; i++ ) {
       vec->info[i] = (void *)malloc(sizeof(int));
       vec->len[i] = (unsigned long) sizeof(int);
    }

    /* pass the completion handler back by reference */
    *completion_handler = compl_hndlr;

    return vec;
}


int main( int argc, char **argv )
{

lapi_handle_t handle;                   /* LAPI handle */
lapi_info_t   info;                     /* Info to pass to LAPI_Init */
int           task_id;                  /* Our LAPI Task id */
int           num_tasks;                /* Total number of tasks */
int           i;                        /* Loop counter */
int           buddy;                    /* Our communication partner */
int           val1, val2;               /* Needed for waitcntr calls on source */

    /* Clear the struct.  Not passing any info to init through this struct. */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks < 2) || ((num_tasks % 2) != 0) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n",
num_tasks);
        exit(1);
    }
 
    /* get address of header handler on tgt */
    hdr_hndlr_list = (void **)malloc(num_tasks*(sizeof(void *)));
    CHECK((LAPI_Address_init(handle,(void *)header_handler,
                                         hdr_hndlr_list)));

    /*
    ** up to this point, all instructions have executed on all tasks.
    ** we now begin differentiating tasks.
    */

    if ((task_id %2) == 0) {   /* sender */
        lapi_vec_t *vec;       /* origin vector */

        /* setup buddy pairs as (0,1), (2,3), etc. */
        buddy = task_id+1;

        /* initialize origin counter */
        CHECK((LAPI_Setcntr(handle,&origin_cntr,0)));

        /* initialize origin vector */
        vec = (lapi_vec_t *) malloc(sizeof(lapi_vec_t));
        vec->len = (unsigned long *)malloc(NUM_VECS*sizeof(unsigned long));
        vec->info = (void **) malloc(NUM_VECS*sizeof(void *));
        vec->num_vecs = NUM_VECS;
        vec->vec_type = LAPI_GEN_IOVECTOR;
        for( i=0; i < NUM_VECS; i++ ) {
            int *int_ptr;
            vec->info[i] = (void *)malloc(sizeof(int));
            int_ptr = (int *)(vec->info[i]);
            *int_ptr = i;
            vec->len[i] = (unsigned long) sizeof(int);
        }

        /*
        ** initialize the completion counter. When it increments, we know
        ** that the completion handler has returned
        */
        CHECK((LAPI_Setcntr(handle,&compl_cntr,0)));

        /* collective call.  sync before data transfer. */
        CHECK((LAPI_Gfence(handle)));

    /*
    ** execute the data transfer to our buddy task.  send ARRAYLEN ints, starting
    ** with data_buffer[0].  Note that no user header is sent, and no counters are
    ** used.
    ** args:  handle                 -- the LAPI handle
    **        buddy                  -- the target task id
    **        hdr_hndlr_list[buddy]  -- the header handler that will execute on
    **                                  the target
    **        NULL                   -- the user header
    **        0                      -- the user header length
    **        vec                    -- the origin vector 
    **        NULL                   -- target counter
    **        origin_cntr            -- origin counter
    **        compl_cntr             -- completion counter
    */
        CHECK((LAPI_Amsendv(handle, buddy, (void *)hdr_hndlr_list[buddy],
                                        NULL, 0, vec, NULL, &origin_cntr, &compl_cntr)));

        /*
        ** wait on origin counter.  once incremented, origin vector is free to be reused
        ** (or in this case freed).
        */
        CHECK((LAPI_Waitcntr(handle,&origin_cntr,1,&val1)));

        /* free all origin vector structs */
        for( i=0; i < NUM_VECS; i++ ) {
           free(vec->info[i]);
        }
        free(vec->info);
        free(vec->len);
        free(vec);

        /* wait for message completion before final termination */
        CHECK((LAPI_Waitcntr(handle, &compl_cntr,1, &val2)));

    } else {             /* receiver */
    /* not used.  but provided for explanation. */
        buddy = task_id-1;

        /* to match the src's sync before data transfer */
        CHECK((LAPI_Gfence(handle)));
    }

    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(hdr_hndlr_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


