

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

#define VECLEN         15
#define NUM_VECS       10
#define ARRAYLEN       10

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];



/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


#define PAD 3
#define NUMCHARS 5

/*
** Example Program to illustrate the LAPI_GEN_STRIDED_XFER type
** using the LAPI_Amsendv api call.
** For a set of n tasks 0,1,...,n-1, where n is an even number, all
** tasks are divided into (src,tgt) buddy pairs (0,1), (2,3), etc.
** For each pair, a vector of strings of strided data is transferred
** from src to tgt with a single LAPI_Amsendv call.  Each vector
** contains  a single int.  The receive completion handler verifies the
** results of the Amsend call by printing the contents of the
** target-side data buffer.  On the source side, the completion counter
** is used to ensure execution of the target completion handler before
** job termination.
*/

/* stores initial value on src and tgt and final value on tgt */
void        *data_buffer;

/* origin vector on src, target vector on tgt */
lapi_vec_t *vec;

/* list of header handler addrs */
void **hdr_hndlr_list;

/* updates on src at msg completion */
lapi_cntr_t   compl_cntr;

/* updates on src when buffer is available */
lapi_cntr_t   org_cntr;

/* completion handler on target */
void compl_hndlr(lapi_handle_t *handle, void *param)
{
    int i;               /* loop counter */
    uint len;            /* data length */

    printf("Greetings from the tgt completion handler...\n");

    /* pull len from user passed param */
    len = (uint) param;

    /* print results */
    for( i = 0; i < len; i++ ) {
        printf("final buffer[%d]: %c\n",i, ((char *)data_buffer)[i]);
    }

    /* free structures allocated in header handler */
    free( data_buffer );
    free( vec->info );
    free( vec );
}

/* header handler on target */
lapi_vec_t *header_handler(lapi_handle_t *handle, void *uhdr,
                               uint uhdr_len, uint vec_len,
                               compl_hndlr_t **completion_handler,
                               void **user_info) 
{
    int i,k;                   /* loop counters */
    int block_size;            /* block size */
    int data_size;             /* stride */

    printf("Greetings from the tgt header handler...\n");
    /* 
    ** setup target vector to match that on the src
    ** compl_hndlr will free.
    */ 
    vec = (lapi_vec_t *)malloc(sizeof(lapi_vec_t));
    vec->len = (unsigned long *)malloc(NUM_VECS*sizeof(unsigned long));
    vec->info = (void **) malloc(3*sizeof(void *));
    vec->num_vecs = NUM_VECS;
    vec->vec_type = LAPI_GEN_STRIDED_XFER;

    block_size = NUMCHARS;
    data_size = (NUM_VECS*block_size) +
                ((NUM_VECS-1)*PAD);
    vec->info[1] = (void *) (block_size);
    vec->info[2] = (void *) (block_size + PAD);

    /* set data buffer pointer as address for data copy */
    data_buffer = (void *) malloc(data_size);
    bzero(data_buffer,data_size);
    vec->info[0] = data_buffer;

    /* set compl hndlr and param */
    *completion_handler = compl_hndlr;
    *user_info = (void *)data_size;

    return vec;
}

/* populate a vector with data */
void fill_vector ( lapi_vec_t* lapi_vec )
{

  int i, j;                    /* loop counters */
  int total_length;
  int data_size;
  int block_size;
  char *pos_ptr;               /* position holder in vec */

  static char digits[] = 
    { "0123456789" };
   
    pos_ptr = lapi_vec->info[0];

    block_size  = (int) lapi_vec->info[1];
    data_size   = (int) lapi_vec->info[2];

    for ( i = 0; i < lapi_vec->num_vecs; i++){
        for ( j = 1; j <= block_size; j++ ){
            *pos_ptr = digits[(j-1+i)%10];
          pos_ptr++;
        }

        /* Store '$' in the pad space */
        if ( (i < lapi_vec->num_vecs-1) && (lapi_vec->num_vecs > 1)) 
         for ( j = 1; j <= data_size-block_size; j++ ){
          *pos_ptr = '$';  
          pos_ptr++;
          }
    }
}

int main( int argc, char **argv )
{

lapi_handle_t handle;                      /* LAPI handle */
lapi_info_t   info;                        /* to pass to LAPI_Init */
int           task_id;                     /* Our LAPI Task id */
int           num_tasks;                   /* Total number of tasks */
int           i,j;                         /* Loop counter */
int           buddy;                       /* Communication partner */
int           val1, val2;                  /* For waitcntr calls */
int           data_size;                   /* Total size of data */
int           *data_ptr;
int           block_size;                  /* Block transfer size */

    /*
    ** Clear the struct.  Not passing any info to init through this
    ** struct.
    */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks % 2) != 0 ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }

    /* get address of header handler on tgt */
    hdr_hndlr_list = (void **)malloc(num_tasks*(sizeof(void *)));
    CHECK((LAPI_Address_init(handle,(void *)header_handler,
                                         hdr_hndlr_list)));

    /*
    ** up to this point, all instructions have executed on all tasks.
    ** we now begin differentiating tasks.
    */

    if ((task_id %2) == 0) {   /* sender */

        /* setup buddy pairs as (0,1), (2,3), etc. */
        buddy = task_id+1;

        /* setup vector for strided transfer */ 
        vec = (lapi_vec_t *) malloc(sizeof(lapi_vec_t));
        vec->info = (void **) malloc(3*sizeof(void *));
        vec->num_vecs = NUM_VECS;
        vec->vec_type = LAPI_GEN_STRIDED_XFER;

        /* block_size is the actual size of data blocks */
        block_size = NUMCHARS;

        /* data size is the total amount of data to be transferred */
        data_size = (NUM_VECS*block_size) +
                    ((NUM_VECS-1)*PAD);

        /* info[0] is the base address of the data to transfer. */
        vec->info[0] = (void *) malloc(data_size);

        /* info[1] is the block size */
        vec->info[1] = (void *) (block_size);

        /* info[2] is the data stride */
        vec->info[2] = (void *) (block_size + PAD);

        /* use helper routine to fill info[0] with data */
        fill_vector(vec);

        /* dump contents after fill */
        for( i = 0; i < data_size; i++ ) {
            printf("vec->info[0][%d]: %c\n",i,
                    ((char *)((vec->info)[0]))[i]);
        } 
 
        /* 
	** initialize origin counter.  When it increments, we know
        ** that the data buffer on the origin sid eis available
        */
        CHECK((LAPI_Setcntr(handle,&org_cntr,0)));
 
       /*
       ** initialize the completion counter. When it increments, we know
       ** that the completion handler has returned
       */
       CHECK((LAPI_Setcntr(handle,&compl_cntr,0)));


       /* sync before transfer */
       CHECK((LAPI_Gfence(handle)));

    /*
    ** execute the strided vector transfer to our buddy task.
    ** Note that no user header is sent and no counters are used.
    ** args:  handle                 -- the LAPI handle
    **        buddy                  -- the target task id
    **        hdr_hndlr_list[buddy]  -- target header handler
    **        NULL                   -- the user header
    **        0                      -- the user header length
    **        vec                    -- the origin vector
    **        NULL                   -- target counter
    **        org_cntr               -- origin counter
    **        compl_cntr             -- completion counter
    */ 
       CHECK((LAPI_Amsendv(handle, buddy, 
                          (void *)(hdr_hndlr_list[buddy]), NULL, 
                           0,vec, NULL, &org_cntr, &compl_cntr)));

        /*
        ** wait on origin counter.  once incremented, origin vector is
        ** free to be reused (or in this case freed).
        */
        CHECK((LAPI_Waitcntr(handle,&org_cntr,1,&val1)));

        /* free all origin vector structs */
        free(vec->info[0]);
        free(vec->info);
        free(vec);

        /* wait for message completion before final termination */
        CHECK((LAPI_Waitcntr(handle, &compl_cntr,1, &val2)));

    } else {             /* receiver */

        /* match src's sync */
        CHECK((LAPI_Gfence(handle)));
    }

    /* the remaining code will be executed by all tasks */

    /* final sync before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(hdr_hndlr_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;

}


