

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Example Program to illustrate the use of the LAPI_Xfer interface
**   to the Amsend api call.
**   For a set of n tasks 0,1,...,n-1, where n is an even number, all
**   tasks are divided into (src,tgt) buddy pairs (0,1), (2,3), etc.
**   For each pair, an array of ints is transferred from src to tgt
**   with a single LAPI_Amsend call.  The receive completion handler
**   verifies the results of the Amsend call by printing the contents
**   of the target-side data buffer.  LAPI will automatically
**   invoke the completion handler upon message completion.  Note that
**   the origin task must wait on the completion counter before 
**   terminating, to ensure that the completion handler has executed.
*/

/* stores initial value on src and final value on tgt */
int           data_buffer[ARRAYLEN];

/* global array of header handler addrs */
lapi_long_t   *hdr_hdl_list;

/* updates on src at msg completion */
lapi_cntr_t   compl_cntr;


/*
**  completion handler that runs on target after completion of message
**  delivery. simply prints the contents of the data buffer that is the
**  target of the Amsend call
*/
void compl_hndlr(lapi_handle_t *handle, void *param)
{
    int i;              /* loop counter */
    
    printf("Greetings from the completion handler...\n");
    for( i = 0; i < ARRAYLEN; i++ ) {
        printf("final buffer[%d]: %d\n",i, data_buffer[i]);
    }

}

/* header handler that runs on target when first packet arrives.
**   sets the completion_handler pointer and returns the address
**   of the data buffer for message delivery.  LAPI writes the
**   Amsend data at this address.
*/
void *header_handler(lapi_handle_t *handle, void *uhdr, uint uhdr_len,
                     uint msg_len, compl_hndlr_t **completion_handler,
                     void **user_info) {

    /* pass the completion handler back by reference */
    *completion_handler = compl_hndlr; 
    printf("Greetings from the header_handler, returning:0x%x\n",data_buffer);
    return data_buffer;
}


int main( int argc, char **argv )
{

lapi_handle_t handle;                      /* LAPI handle */
lapi_info_t   info;                        /* Info to pass to LAPI_Init */
int           task_id;                     /* Our LAPI Task id */ 
int           num_tasks;                   /* Total number of tasks */
void          *buf_addr_list[NUM_TASKS];   /* Table of remote data buf addrs. */
int           i,k;                         /* Loop counter */
int           buddy;                       /* Our communication partner */
int           val;                         /* Needed for waitcntr call */
lapi_xfer_t   xfer_struct;                 /* Data structure for the xfer call */

    /* Clear the struct.  Not passing any info to init through this struct. */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id))); 

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks % 2) != 0 || (num_tasks < 2) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }  
    hdr_hdl_list = (lapi_long_t *) malloc(num_tasks*sizeof(lapi_long_t));

    /*
    ** collective calls.  each task stores its own addr plus remote
    ** addrs
    */
    CHECK((LAPI_Address_init64(handle,(lapi_long_t)&header_handler,
                       hdr_hdl_list)));
    /*
    ** up to this point, all instructions have executed on all tasks.
    ** we now begin differentiating tasks.
    */
    if ((task_id %2) == 0) {   /* message source */

    /* setup buddy pairs as (0,1), (2,3), etc. */
        buddy = task_id+1;

    /* initialize data buffer. */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = i*buddy;
        }

    /* collective call.  sync before starting data transfer */
        CHECK((LAPI_Gfence(handle)));

    /*
    ** initialize the completion counter. When it increments, we know
    ** that the completion handler has returned
    */
    CHECK((LAPI_Setcntr(handle,&compl_cntr,0)));

    /* 
    ** Setup the xfer struct to transfer to our buddy task.  send ARRAYLEN ints,
    ** starting with data_buffer[0].  Note that no user header is sent, 
    ** and no counters are used.
    ** args:  handle                 -- the LAPI handle
    **        buddy                  -- the target task id
    **        header_handler         -- the header handler that will 
    **                                  execute on the target
    **        NULL                   -- the user header 
    **        0                      -- the user header length
    **        &(data_buffer[0])      -- the starting address of data to
    **                                  transfer
    **        ARRAYLEN*sizeof(int)   -- the length of data to transfer
    **        NULL, NULL             -- target and origin counters
    **        &compl_cntr            -- completion counter
    */
        xfer_struct.Am.Xfer_type = LAPI_AM_XFER;
        xfer_struct.Am.flags = 0;
        xfer_struct.Am.tgt = buddy;
        xfer_struct.Am.hdr_hdl = (lapi_long_t) (hdr_hdl_list[buddy]);
        xfer_struct.Am.uhdr = (void *) NULL;
        xfer_struct.Am.uhdr_len = 0;
        xfer_struct.Am.udata = (void *) data_buffer;
        xfer_struct.Am.udata_len = ARRAYLEN*(sizeof(int));
        xfer_struct.Am.shdlr = (scompl_hndlr_t *) NULL;
        xfer_struct.Am.sinfo = (void *) 0;
        xfer_struct.Am.org_cntr = NULL;
        xfer_struct.Am.cmpl_cntr = &compl_cntr;
        xfer_struct.Am.tgt_cntr = (lapi_long_t) NULL;
        CHECK((LAPI_Xfer(handle, &xfer_struct)));

        /* wait for message completion before final termination */
        CHECK((LAPI_Waitcntr(handle, &compl_cntr,1, &val)));


     } else {             /* receiver */
         /* buddy = task_id-1; */

         /* initialize the data buffer (not necessary, but provided for
            clarity) */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = 0;
        }

        /* to match the src's sync before data transfer */
        CHECK((LAPI_Gfence(handle)));
    }

    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(hdr_hdl_list);
    
    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


