

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Example Program to illustrate the use of the LAPI_Xfer interface
**   to the Put api call.
**   For a set of n tasks 0,1,...,n-1, where n is an even number, all
**   tasks are divided into (src,tgt) buddy pairs (0,1), (2,3), etc.
**   For each pair, an array of ints is transferred from src to tgt
**   with a single LAPI_Put call.  The target counter is used for
**   synchronization.  Once the message completes, the target shows
**   the result by printing the contents of the data buffer.
*/


/* stores initial value on src and final value on tgt */
int           data_buffer[ARRAYLEN];

/* store list of remote buffer addrs */
lapi_long_t   *data_buffer_list;  /* remote data buffer addrs */

/* store list of remote tgt_cntr addrs */
lapi_long_t   *tgt_cntr_list;     /* remote tgt cntr addrs    */

/* updates on tgt at msg completion */
lapi_cntr_t   tgt_cntr;


int main( int argc, char **argv )
{
lapi_handle_t handle;        /* LAPI handle */
lapi_info_t   info;          /* Info to pass to LAPI_Init */
int           task_id;       /* Our LAPI Task id */
int           num_tasks;     /* Total number of tasks */
int           i,j;           /* Loop counters */
int           buddy;         /* Our communication partner */
int           val;           /* Needed for waitcntr call on source */
lapi_xfer_t   xfer_struct;   /* Data structure for the xfer call */

    /* Not passing any info to init through this struct. */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks % 2) != 0 || (num_tasks < 2) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }

    data_buffer_list = (lapi_long_t *) malloc(num_tasks*sizeof(lapi_long_t));
    tgt_cntr_list = (lapi_long_t *) malloc(num_tasks*sizeof(lapi_long_t));

    /* each task executes on tgt_cntr within its own data space */
    CHECK((LAPI_Setcntr(handle,&tgt_cntr,0)));

    /*
    ** collective calls.  each task stores its own addr plus remote
    ** addrs
    */
    CHECK((LAPI_Address_init64(handle,(lapi_long_t)&tgt_cntr,tgt_cntr_list)));
    CHECK((LAPI_Address_init64(handle,(lapi_long_t)data_buffer,data_buffer_list)));

    /*
    ** up to this point, all instructions have executed on all tasks.
    ** we now begin differentiating tasks.
    */
    if ( task_id % 2  == 0 ){   /* sender */
        buddy = task_id + 1;

        /* initialize the data buffer */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = i*(1+task_id);
        }

        /* collective call.  sync before starting data transfer */
        CHECK((LAPI_Gfence(handle)));

        /*
        ** execute the data transfer to our buddy task.  send ARRAYLEN
        ** ints, starting with data_buffer[0].  Note that the target counter
        ** is used to signal message completion at the target.
        ** args:  handle                  -- the LAPI handle
        **        LAPI_PUT_XFER           -- indicates xfer type to LAPI
        **        buddy                   -- the target task id
        **        flags                   -- not used.  cleared to avoid
        **                                   any errant data
        **        data_buffer_list[buddy] -- remote addr for writing
        **                                   data
        **        &(data_buffer[0])       -- the starting address of
        **                                   data to transfer
        **        ARRAYLEN*sizeof(int)    -- the length of data to
        **                                   transfer
        **        NULL, NULL              -- send completion handler
        **                                   and send compl handler info
        **                                   (neither is used)
        **        NULL, NULL              -- origin and compl counters
        **        tgt_cntr_list[buddy]    -- target counter.  will update
        **                                   when message transfer completes
        **                                   at the target
        */
        xfer_struct.Put.Xfer_type = LAPI_PUT_XFER;
        xfer_struct.Put.tgt = buddy;
        xfer_struct.Put.flags = 0;
        xfer_struct.Put.tgt_addr = data_buffer_list[buddy];
        xfer_struct.Put.org_addr = (void *) &(data_buffer[0]);
        xfer_struct.Put.len = (ulong) ARRAYLEN*(sizeof(int));
        xfer_struct.Put.shdlr = (scompl_hndlr_t *) NULL;
        xfer_struct.Put.sinfo = (void *) NULL;
        xfer_struct.Put.org_cntr = NULL;
        xfer_struct.Put.cmpl_cntr = NULL;
        xfer_struct.Put.tgt_cntr = (lapi_long_t) tgt_cntr_list[buddy];
        CHECK((LAPI_Xfer(handle, &xfer_struct)));
    } else {             /* receiver */
        buddy = task_id - 1;

        /* to match the src's sync before data transfer */
        CHECK((LAPI_Gfence(handle)));

        /* wait for data to arrive from src */
        CHECK((LAPI_Waitcntr(handle, &tgt_cntr,1, &val)));

        printf("Received data from buddy %d\n",buddy);
        for (i = 0; i < ARRAYLEN; i++ ) {
            printf("data_buffer[%d]: %d\n", i, data_buffer[i]);
        }
    }

    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(data_buffer_list);
    free(tgt_cntr_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


