/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 2008               */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
#pragma comment(copyright,"5765-543 (C) Copyright IBM Corp. 2008 @(#) 1.2 src/ppe/hpct/source/examples/plugins/average.c, ppe.hpct, ppe_rhya 08/12/01 14:19:38")
/****************************************************************************
 *
 *   This plugin implements two functionalities.
 *   - The distributor distributes the groups in a round robin fashion 
 *     among the MPI tasks.
 *   - The aggregator takes the average of all values from the same group.
 *     The aggregator detects the group by itself. Hence it would also
 *     cooperate with a different distributor.
 *
 *               C. Pospiech ** ACTC
 *
 *  $Id: average.c,v 1.4 2008/02/05 11:43:20 cp Exp $
 *
 ****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <mpi.h>

#include "hpc_data.h"
#include "hpc_agg.h"

#define TRUE_ 1
#define FALSE_ 0

#define PRINT_TASK 0
#define MAX_GROUPSTR 50


int distributor(void) {

  int my_process_id = MPI_TASK_ID_NOT_AVAILABE;
  long *group;
  int index, read_ctr = 0;
  size_t group_size = 0;
  char *envstr = NULL, *endptr = NULL, groupstr[MAX_GROUPSTR];

  /*
   * If we want to distribute different HPM group settings 
   * to different tasks (in a round robin fashion) we need to
   * get hold of the MPI task id that the current process is running on.
   *
   * As indicated in the documentation, the distributor must not call
   * any MPI functions. Furtunately the MPI task ID can be extracted as an
   * environment variable in many cases.
   */

  /* 
   * On AIX, if MP_CHILD is set,
   * the process is running under poe and this variable is the MPI task id.
   */
  envstr = getenv( "MP_CHILD" );
  if ( envstr ) {
    my_process_id = atol(envstr);
  }

  /* 
   * On MVAPICH, if MPIRUN_RANK is set,
   * the process is running under mpirun and this variable is the MPI task id.
   */
  envstr = getenv( "MPIRUN_RANK" );
  if ( envstr ) {
    my_process_id = atol(envstr);
  }

  /* 
   * On OpenMPI, if OMPI_MCA_ns_nds_vpid is set,
   * the process is running under mpirun and this variable is the MPI task id.
   */
  envstr = getenv( "OMPI_MCA_ns_nds_vpid" );
  if ( envstr ) {
    my_process_id = atol(envstr);
  }

  /*
   * On Myrinet MPICH/MX-GM, if GMPI_ID is set,
   * the process is running under mpirun and this variable is the MPI task id.
   */
  envstr = getenv( "GMPI_ID" );
  if ( envstr ) {
    my_process_id = atol(envstr);
  }

  /*
   * On Intel MPI, if PMPI_RANK is set,
   * the process is running under mpirun and this variable is the MPI task id.
   */
  envstr = getenv( "PMPI_RANK" );
  if ( envstr ) {
    my_process_id = atol(envstr);
  }

  /*
   * On MPICH, the environment variable GMPI_OPTS
   * contains both, the task id and the number of tasks.
   * GMPI_OPTS=m<task_id>n<num_tasks>
   */
  envstr = getenv( "GMPI_OPTS" );
  if ( envstr ) {
    my_process_id = strtol(envstr+1,&endptr, 10);
  }

  /*
   * If my_process_id could not be determined return without action
   * and without error message.
   */
  if ( my_process_id == MPI_TASK_ID_NOT_AVAILABE) return(0);

  /*
   * The distributor is only supposed to manipulate environment variables.
   *
   * Let's read the variable HPM_EVENT_SET.
   */
  envstr = getenv( "HPM_EVENT_SET" );
  /*
   * If HPM_EVENT_SET is not set, there is nothing to distribute.
   * Just exit in this case.
   */
  if ( !envstr ) return(0);
  /*
   * The following loop reads the environment variable and (re)allocates
   * the vector of HPM group values on the fly.
   */

  endptr = envstr;
  while ( endptr && ( *endptr != '\0' )) {
    if (read_ctr >= group_size ) {
      group_size = (group_size == 0) ? 16 : group_size*2;
      /* 
       * if realloc goes wrong, we return the value of errno and
       * rely on the error handling of HPM to turn that into a readable
       * error message.
       */
      if ((group = (long *) realloc(&group, group_size)) == NULL) return errno;
    }
    group[read_ctr++] = strtol(endptr, &endptr, 10);
    if ( errno ) return errno;
    /* progress the pointer beyond the separator */
    if ( endptr && (*endptr != '\0')) endptr++;
  }

  /*
   * Now read_ctr actually contains the number of elements that could be
   * read from HPM_EVENT_SET.
   *
   * Round robin just means that we have to take the modulus with respect
   * read_ctr. If read_ctr is 1 (or less) there is again not much 
   * to distribute.
   */
  if ( read_ctr <= 1 ) return(0);
  index = my_process_id % read_ctr; /* This implies index < read_ctr */
  sprintf(groupstr,"%d",group[index]);
  if (setenv("HPM_EVENT_SET",groupstr,1) != 0) return errno;

  /* This is the normal exit - HPM_EVENT_SET has been redefined */
  return(0);
}

int aggregator(int num_in, hpm_event_vector in,
               int *num_out, hpm_event_vector *out,
	       int *is_print_task) {

  int MPI_is_initialized, MPI_Init_done_here;
  int argc=1;
  char *dummy_string = "dummy String";
  char *dummy_vector[1] = {dummy_string};
  char **argv = dummy_vector;

  hpm_event_vector new_vector, old_vector = *out;
  hpm_event_info *new;
  int i, j, ct, num_values=0, rc=0;
  int key, color=0, mask=(1<<20)-1, my_id, my_size;
  MPI_Comm aggr_comm;
  double *in_times, *out_times;
  double *in_values, *out_values;
  double my_size_inverse;


  /* 
   * Initialize MPI
   * Depending on whether this function is called from hpmcount
   * or libhpm, MPI_Init might not be needed.
   * So we better check whether MPI has already been initialized.
   */
  if (MPI_Initialized( &MPI_is_initialized ) != MPI_SUCCESS) {
    printf("Error at MPI_Initialized\n");
    MPI_Abort(MPI_COMM_WORLD,4);
  }
  MPI_Init_done_here = !MPI_is_initialized;
  if (MPI_Init_done_here && (MPI_Init(&argc, &argv) != MPI_SUCCESS)) {
    printf("Error at MPI_Init\n");
    MPI_Abort(MPI_COMM_WORLD,4);
  }

  /*
   * As a first step we create a MPI communicator which splits
   * MPI_COMM_WORLD into groups along which aggregation makes sense,
   * i.e. MPI tasks counting the same groups.
   *
   * To feed this information into a call to MPI_Comm_split we cook
   * up a value color to represent some kind of hash value of the groups
   * that are counted on this processor. This means that color is the HPM
   * group number, if there is only one group counted on this MPI task.
   *
   * The key is taken to be the MPI task id with respect to MPI_COMM_WORLD
   */
  for (i=0; i<num_in; i++) {
    color = ((color << 7) & mask) + in[i].data[0].group_id;
  }
  if (MPI_Comm_rank(MPI_COMM_WORLD, &key) != MPI_SUCCESS) {
    printf("Error at MPI_Comm_rank\n");
    MPI_Abort(MPI_COMM_WORLD,4);
  }
  if (MPI_Comm_split(MPI_COMM_WORLD, color, key, &aggr_comm) != MPI_SUCCESS) {
    printf("Error at MPI_Comm_split\n");
    MPI_Abort(MPI_COMM_WORLD,4);
  }
  if (MPI_Comm_rank(aggr_comm, &my_id) != MPI_SUCCESS) {
    printf("Error at MPI_Comm_rank\n");
    MPI_Abort(MPI_COMM_WORLD,4);
  }
  if (MPI_Comm_size(aggr_comm, &my_size) != MPI_SUCCESS) {
    printf("Error at MPI_Comm_size\n");
    MPI_Abort(MPI_COMM_WORLD,4);
  }
  my_size_inverse = 1.0/my_size;

  /*
   * Next the necessary send and receive buffers are allocated.
   * We start by calulating the length of the value buffers.
   */
  for (i=0; i<num_in; i++) num_values += in[i].num_data;
  if ((in_times = (double*) malloc(num_in*HPM_NTIM*sizeof(double))) == NULL) 
    return errno;
  if ((out_times = (double*) malloc(num_in*HPM_NTIM*sizeof(double))) == NULL)
    return errno;
  if ((in_values = (double*) malloc(num_values*sizeof(double))) == NULL)
    return errno;
  if ((out_values = (double*) malloc(num_values*sizeof(double))) == NULL)
    return errno;
  if ((  new_vector = (hpm_event_vector)
	 realloc(old_vector, num_in*sizeof(hpm_data_item))) == NULL) {
    if (errno) rc = errno;
    else rc = -1;
    fprintf(stderr,"Error in allocating the output vector.\n");
    fprintf(stderr,"Size requested %d\n",num_in*sizeof(hpm_data_item));
    fprintf(stderr,"old_vector points to %p\n", old_vector);
    return(rc);
  }

  /*
   * Populate the time buffers and start computing averages.
   */
  ct = 0;
  for (i=0; i<num_in; i++)
    for (j=0; j<HPM_NTIM; j++)
      in_times[ct++] = in[i].times[j];
  if (MPI_Allreduce(in_times, out_times, ct, 
		    MPI_DOUBLE, MPI_SUM, aggr_comm) != MPI_SUCCESS) {
    printf("Error at MPI_Reduce\n");
    MPI_Abort(MPI_COMM_WORLD,4);
  }
  ct = 0;
  for (i=0; i<num_in; i++)
    for (j=0; j<HPM_NTIM; j++)
      new_vector[i].times[j] = out_times[ct++] * my_size_inverse;

  /*
   * Now Populate the value buffers. Then the average is taken.
   * All computations are carried out in double to minimize round off
   * errors.
   */
  ct = 0;
  for (i=0; i<num_in; i++)
    for (j=0; j<in[i].num_data; j++) {
      in_values[ct++] = in[i].data[j].value;
      /* 
       * If a non-zero wallclock time is provided for a value, 
       * we could think of normalizing by
       * the ratio of actual time versus averaged time.
      if ( in[i].times[HPM_TIME_WALLCLOCK] != 0 )
	in_values[ct-1] *= (new_vector[i].times[HPM_TIME_WALLCLOCK]
			    / in[i].times[HPM_TIME_WALLCLOCK]);
      */
    }
  if (MPI_Allreduce(in_values, out_values, ct, 
		    MPI_DOUBLE, MPI_SUM, aggr_comm) != MPI_SUCCESS) {
    printf("Error at MPI_Reduce\n");
    MPI_Abort(MPI_COMM_WORLD,4);
  }

  /*
   * Finally, fill the result vector - except for the time values
   * which have been computed above.
   */
  ct = 0;
  for (i=0; i<num_in; i++) {
    if ((new = (hpm_event_info*) 
	 malloc(in[i].num_data*sizeof(hpm_event_info)) ) == NULL ) 
      return errno;
    /*
     * Make a complete copy of the data vector -- including the pointers.
     * This uses the fact that everything happens locally
     * on the same MPI task. Thus the pointers point 
     * to strings in local memory.
     */
    memcpy(new, in[i].data, in[i].num_data*sizeof(hpm_event_info));
    /* Now plug in the new average values */
    for (j=0; j<in[i].num_data; j++)
      new[j].value = out_values[ct++] * my_size_inverse;
    /* Finally copy the remaining values of the hpm_data_item */
    new_vector[i].num_data       = in[i].num_data;
    new_vector[i].data           = new;
    new_vector[i].is_mplex_cont  = in[i].is_mplex_cont;
    new_vector[i].is_rusage      = in[i].is_rusage;
    new_vector[i].mpi_task_id    = key;
    new_vector[i].instr_id       = in[i].instr_id;
    new_vector[i].count          = in[i].count;
    new_vector[i].is_exclusive   = in[i].is_exclusive;
    new_vector[i].xml_element_id = in[i].xml_element_id;
    new_vector[i].description    = in[i].description;
    new_vector[i].xml_descr      = in[i].xml_descr;
  }

  /* Now set the output variables */
  *num_out = num_in;
  *out = new_vector;
  *is_print_task = (my_id == PRINT_TASK);

  /* Free intermediate buffers */
  free(in_times);
  free(out_times);
  free(in_values);
  free(out_values);

  /* ...and exit */
  if (MPI_Init_done_here) MPI_Finalize();

  return(0);
}
